//============================================================================
// Name        : IndependentSet.cpp
// Author      : Bergman, Cire
// Version     :
// Copyright   : 
//============================================================================

#include <iostream>

#include "core/indepset_solver.hpp"
#include "core/orderings.hpp"
#include "core/merge.hpp"
#include "instance/instance.hpp"
#include "util/stats.hpp"

using namespace std;

int main(int argc, char* argv[]) {

  if( argc < 5 ) {
    cout << "\nUsage: indepset [graph-file] [bdd-width] [ordering] [merge] [opt.: ordering parameters]\n\n";
    cout << "\twhere: \n";
    cout << "\t\tgraph-file: graph in DIMACS format\n";
    cout << "\t\tbdd-width: maximum BDD width\n";
    cout << endl;
    cout << "\t\tordering: vertex ordering, options are:\n";
    cout << "\t\t\t1: random\n";
    cout << "\t\t\t2: maximal path decomposition\n";
    cout << "\t\t\t3: min in-state heuristic\n";
    cout << "\t\t\t4: cut vertex decomposition for trees\n";
    cout << "\t\t\t5: cut vertex decomposition for general graphs\n";
    cout << "\t\t\t6: randomized min in-state heuristic (requires probability parameter p)\n";
    cout << "\t\t\t7: fixed parameter (requires ordering file)\n";
    cout << "\t\t\t8: min vertex degree\n";
    cout << endl;
    cout << "\t\tmerge: merging technique, options are:\n";
    cout << "\t\t\t1: all minimum longest path nodes\n";
    cout << "\t\t\t2: pair-by-pair by longest path\n";
    cout << "\t\t\t3: consecutive pairs by longest path\n";
    cout << "\t\t\t4: smallest size\n";
    cout << "\t\t\t5: largest size\n";
    cout << "\t\t\t6: lexicographic ordering\n";
    cout << "\t\t\t7: symmetric difference\n";
    cout << "\t\t\t8: random\n";
    cout << endl;
    cout << endl;
    exit(1);
  }

  Stats stats;
  stats.register_name("time-bdd");

  // read independent set instance
  IndepSetInst* inst = new IndepSetInst;
  inst->read_DIMACS(argv[1]);
  int width = atoi(argv[2]);

  //	  ofstream out_degree("out-degree.txt", ios::app);
  //	  out_degree << argv[1] << " " << inst->graph->n_vertices << " " << inst->graph->n_edges << " ";
  //	  double density = 2*((double)inst->graph->n_edges) / ((double)inst->graph->n_vertices * (inst->graph->n_vertices-1) );
  //	  out_degree << density;
  //	  out_degree << endl;
  //	  out_degree.close();
  //	  return 0;

  cout << "\n\n*** Independent set - Bounds ***" << endl;
  cout << "\twidth: " << width << endl;


  stats.start_timer(0);

  // read ordering type
  int order_n = atoi(argv[3]);
  IS_Ordering* ordering = NULL;
  switch( order_n ) {
  case 1:
    {
      //		// extract instance identifier
      //		int id = 0;
      //		char inst_name[256];
      //		strcpy(inst_name, argv[1]);
      //		cout << "random ordering: " << endl;
      //		cout << "\tbroken string: ";
      //
      //		char* name = strtok(inst_name, "-");
      //		cout << name;
      //		name = strtok(NULL, "-");
      //		cout << name << " ";
      //		name = strtok(NULL, "-");
      //		cout << name << " ";
      //		id = atoi(name)*1000;
      //		name = strtok(NULL, ".clq");
      //		cout << name << " ";
      //		id += atoi(name);
      //		cout << endl;
      //		cout << "\trandom seed: " << id << endl;
      //
      //		srand(id);  // initialize seed
      srand(0);
    }
    break;
  case 2:
    ordering = new MaximalPathDecomp(inst);
    break;
  case 3:
    ordering = new MinInState(inst);
    break;
  case 4:
    ordering = new CutVertexDecomposition(inst);
    break;
  case 5:
    ordering = new CutVertexDecompositionGeneralGraph(inst);
    break;
  case 6:
    {
      if( argc < 6 ) {
	cout << "Error: required probability parameter for randomized min_in state\n\n";
	exit(1);
      }

      double prob = atof(argv[5]);
      if( prob < 0 || prob > 1 ) {
	cout << "Error: probability must be between 0 and 1\n\n";
	exit(1);
      }

      ordering = new RandomizedMinInState(inst, prob);
    }
    break;
  case 7:
    {
      if( argc < 6 ) {
	cout << "Error: text file required for ordering\n\n";
	exit(1);
      }
      ordering = new FixedOrdering(inst, argv[5]);
    }
    break;
  case 8:
    {
      ordering = new MinDegreeOrdering(inst);
      break;
    }
  default:
    cout << "\nERROR: ordering undefined\n\n";
    exit(1);
  }


  // read merge type
  int merge_n = atoi(argv[4]);
  IS_Merging* merger = NULL;
  switch( merge_n ) {
  case 1:
    merger = new MinLongestPath(inst, width);
    break;
  case 2:
    merger = new PairMinLongestPath(inst, width);
    break;
  case 3:
    merger = new ConsecutivePairLongestPath(inst, width);
    break;
  case 4:
    merger = new MinSizeMerger(inst, width);
    break;
  case 5:
    merger = new MaxSizeMerger(inst, width);
    break;
  case 6:
    merger = new LexicographicMerger(inst, width);
    break;
  case 7:
    merger = new SymmetricDifferenceMerger(inst, width);
    break;
  case 8:
    {
      srand(time(NULL));
      merger = new RandomMerger(inst, width);
    }
    break;
  default:
    cout << "\nERROR: ordering undefined\n\n";
    exit(1);
  }

  //Bounds bounds;
  int final_width;

  int min = inst->graph->n_vertices+1, max = -1;
  int min_width = INF, max_width = -1;
  double avg = 0, avg_width = 0;

  if( order_n != 1 && merge_n != 8 ) {

    // all orderings except random
    IndepSetSolver solver(inst, width);
    solver.ordering = ordering;
    solver.merger = merger;

    IntSet starting_state;
    starting_state.resize( 0, inst->graph->n_vertices-1, true );

    Bounds bounds;
    bounds.ub = solver.generate_relaxation(starting_state, 0);
    final_width = solver.final_width;
    //bounds.lb = solver.generate_restriction(starting_state, 0);

    avg_width = final_width;
    min_width = final_width;
    max_width = final_width;

    avg = bounds.ub;
    min = bounds.ub;
    max = bounds.ub;

    // write order to file "order.txt"
    ofstream orderfile("order.txt");
    for( int v = 0; v < inst->graph->n_vertices; v++ ) {
      orderfile << solver.vertex_in_layer[v] << " ";
    }
    orderfile.close();

    cout << endl << "Upper bound: " << bounds.ub << 
      " - width: " << final_width << endl;

  } else {

    // random ordering
    IntSet starting_state;
    starting_state.resize( 0, inst->graph->n_vertices-1, true );

    int n_orders = 5;

    for( int i = 0; i < n_orders; i++ ) {

      if( order_n == 1 )
	ordering = new RandomOrdering(inst);

      if( merge_n == 8 )
	merger = new RandomMerger(inst, width);


      IndepSetSolver solver(inst, width);
      solver.ordering = ordering;
      solver.merger = merger;

      int bound = solver.generate_relaxation(starting_state, 0);

      min = MIN(min, bound);
      max = MAX(max, bound);
      avg += bound;

      min_width = MIN(min_width, solver.final_width);
      max_width = MAX(max_width, solver.final_width);
      avg_width += solver.final_width;

      //			if( i != n_orders-1 ) {
      //				delete ordering;
      //			}

      cout << endl << "Upper bound: " << bound << " - width: " << solver.final_width << endl;
    }

    avg /= (double)n_orders;
    avg_width /= (double)n_orders;
  }


  stats.end_timer(0);


  char statfilename[256];
  sprintf(statfilename, "stats.txt");

  ofstream statfile(statfilename, ios::app);
  statfile << argv[1];

  statfile << "\t" << inst->graph->n_vertices;   // instance
  statfile << "\t" << width;
  statfile << "\t" << ordering->name;
  statfile << "\t" << merger->name;

  statfile << "\t" << min_width;  // width
  statfile << "\t" << avg_width;
  statfile << "\t" << max_width;

  statfile << "\t" << min; // bounds
  statfile << "\t" << avg;
  statfile << "\t" << max;

  statfile << "\t" << stats.get_time(0);
  statfile << endl;

  statfile.close();


  return 0;
}
