/*
 * --------------------------------------------------------
 * Instance data structures
 * --------------------------------------------------------
 */

#ifndef INSTANCE_HPP_
#define INSTANCE_HPP_

#include <fstream>
#include <iostream>
#include <cstring>
#include "graph.hpp"

using namespace std;

enum VertexOrder {
  LEX,           /**< lexicographic node ordering */
  CLIQUE,        /**< clique node ordering */
  MIN_CUT,        /**< min. cut node ordering */
  HAM_PATH,       /**< ham path for indep set */
  FEW_VARS_CON    /**< pick the constraint with the fewest number of variables and order them */
};




/**
 * Independent set instance structure
 */
struct SetCoverInst {

	int             					num_vars;
	int             					num_cons;
	int*            					weights;

	vector< vector<int> >     			consts_of_a_var;
	vector< vector<int> >          		vars_in_const;
	vector< vector<bool> >          	is_var_in_const;

//	vector< vector< vector<int> > >     absorption;
//
//	vector< int >             			first_nonempty_abs;
//	int                       			first_first_nonempty_abs;

	void    read_orlib(char* filename);
	void    preprocess();
};


/**
 * Independent set instance structure
 */
struct IndepSetInst {

  Graph*              graph;             /**< independent set graph */
  int*                weights_inclusion; /**< node weights / inclusion */
  int*                weights_exclusion; /**< node weights / exclusion */

  int*			      weights;

  IndepSetInst() : weights(NULL) { }

  /** Read DIMACS independent set instance */
  void read_DIMACS(const char* filename);

  /** Read DIMACS independent set instance with costs */
  int read_DIMACS_with_costs(const char* filename, const char* costs_file, const char* params_file);

  /** Assign weights to vertices */
  void assign_weights();

  /** Assign graph */
  void assign_graph(Graph *graph) {
    if( weights != NULL ) {
      delete[] weights;
    }
    weights = new int[ graph->n_vertices ];
    for( int i = 0; i < graph->n_vertices; ++i ) {
      weights[i] = 1;
    }
    this->graph = graph;
  }

//  /** Reorder graph vertices */
//  void reorder(vector<int> &mapping);

  /** Compute distance variance */
  //long int compute_distance_variance();
};



/*
 * ----------------------------------------
 * Inline implementations
 * ----------------------------------------
 */


/**
 * Read DIMACS independent set instance with no costs
 */
inline void IndepSetInst::read_DIMACS(const char *filename) {

  cout << "Reading instance " << filename << endl;

  // read graph
  graph = new Graph;
  graph->read_dimacs(filename);
  
  cout << "\tnumber of vertices: " << graph->n_vertices << endl;
  cout << "\tnumber of edges: " << graph->n_edges << endl;

  // allocate weights: 1 for inclusion, 0 for exclusion
  weights_inclusion = new int[graph->n_vertices];
  memset(weights_inclusion, -1, sizeof(int)*graph->n_vertices);

  weights_exclusion = new int[graph->n_vertices];
  memset(weights_exclusion, 0, sizeof(int)*graph->n_vertices);

  assign_weights();

  cout << "\tdone." << endl;
}


/**
 * Read DIMACS independent set instance
 */
inline int IndepSetInst::read_DIMACS_with_costs(const char *filename, const char* costs_file, const char* params_file) {

  // read graph
  graph = new Graph;
  graph->read_dimacs(filename);

  cout << "\tnumber of vertices: " << graph->n_vertices << endl;
  cout << "\tnumber of edges: " << graph->n_edges << endl;

  // allocate weights - in this case, they are all 1
  weights_inclusion = new int[graph->n_vertices];
  weights_exclusion = new int[graph->n_vertices];
    
  ifstream costsfile( costs_file );
  for( int i = 0; i < graph->n_vertices; i++ ) {
    costsfile >> weights_inclusion[i];
    costsfile >> weights_exclusion[i];
  }
  costsfile.close();
    
  ifstream paramsfile( params_file );
  int width;
  paramsfile >> width;
    
  vector<int> mapping(graph->n_vertices);
  for( int i = 0; i < graph->n_vertices; i++ ) {
    paramsfile >> mapping[i];
  }
  paramsfile.close();
    
  //reorder(mapping);

  return width;    
}



#endif /* INSTANCE_HPP_ */
