//============================================================================
// Name        : IndependentSet.cpp
// Version     :
// Copyright   : 
//============================================================================


#include <iostream>
#include <ilcplex/ilocplex.h>
#include <list>

#include "instance/instance.hpp"
#include "util/stats.hpp"
#include "util/util.hpp"


using namespace std;


//------------------------------------------------------------
// Global variables for clique decomposition

Graph* original_graph;              // original input graph
vector< vector<int> > cliques;      // cliques in graph
vector< Graph* > clique_graphs;      // graphs representing cliques

//------------------------------------------------------------



//
// Extract clique from graph.
// If no clique can be extracted (which means that
// the graph has no edges), return NULL
//
Graph* extract_clique(Graph* graph) {

  // find vertex with largest degree
  int vertex = -1;
  int max_degree = 0;
  for( int v = 0; v < graph->n_vertices; ++v ) {
    if( graph->degree(v) > max_degree ) {
      vertex = v;
      max_degree = graph->degree(v);
    }
  }

  // if maximum degree is zero, graph has no edges
  if( max_degree == 0 ) {
    return NULL;
  }

  // vertices already in clique (using bool* might be slightly more efficient)
  vector<bool> in_clique( graph->n_vertices, false );
  vector<int> clique;
  clique.push_back( vertex );
  in_clique[vertex] = true;

  // grow clique by taking vertices that are adjacent to 'vertex'
  // with maximum degree
  while ( true ) {

    vector<int> subgraph;
    for( vector<int>::iterator u = graph->adj_list[vertex].begin();
	 u != graph->adj_list[vertex].end();
	 ++u )
      {
	// check if vertex is adjacent to all vertices in clique
	bool is_adj_all = !in_clique[ *u ];
	for( int j = 1; j < (int)clique.size() && is_adj_all; ++j ) {
	  is_adj_all = graph->is_adj( clique[j], *u );
	}
	if( is_adj_all ) {
	  subgraph.push_back( *u );
	}
      }

    if( subgraph.empty() ) {
      break;
    }

    int max_degree = -1;
    int selected_u = -1;
    for( int i = 0; i < (int)subgraph.size(); ++i ) {
      int degree = 0;
      for( int j = 0; j < (int)subgraph.size(); ++j ) {
        if( i == j ) continue;
        if( graph->is_adj( subgraph[i], subgraph[j] ) ) {
          degree++;
        }
      }
      if( degree > max_degree ) {
        selected_u = subgraph[i];
        max_degree = degree;
      }
    }

    // add vertex to clique
    clique.push_back( selected_u );
    in_clique[selected_u] = true;
  }

  // build graph representing clique, and remove clique from original graph
  //Graph* clique_graph = new Graph( graph->n_vertices );
  for( int i = 0; i < (int)clique.size(); ++i ) {
    for( int j = i+1; j < (int)clique.size(); ++j ) {
      //clique_graph->add_edge( clique[i], clique[j] );
      graph->remove_edge( clique[i], clique[j] );
    }
  }

  // add clique to set of cliques
  cliques.push_back( clique );
  //clique_graphs.push_back( clique_graph );

  return original_graph;
}



// ----------------------------------------------------
// Decompose the graph into cliques
//-----------------------------------------------------

void clique_decomposition(
			  IloModel            &model,
			  IloNumVarArray      &x,
			  IndepSetInst*       inst)
{

  // Create original graph
  original_graph = new Graph( inst->graph->n_vertices );
  for( int v = 0; v < inst->graph->n_vertices; ++v ) {
    if( inst->graph->adj_list[v].size() != 0 ) {
      for( vector<int>::iterator it = inst->graph->adj_list[v].begin();
	   it != inst->graph->adj_list[v].end();
	   ++it )
	{
	  original_graph->add_edge( v, *it );
	}
    }
  }


  // Extract cliques
  Graph* graph = inst->graph;
  Graph* clique_graph = extract_clique(graph);

  while( clique_graph != NULL ) {

    //    cout << "Clique " << cliques.size() << ": ";
    //    for( int i = 0; i < (int)cliques.back().size(); ++i ) {
    //      cout << cliques.back()[i] << " ";
    //    }
    //    cout << endl;

    clique_graph = extract_clique(graph);
  }
}


int main(int argc, char* argv[]) {

  Stats stats;
  stats.register_name("time");
  stats.register_name("clique");

  if ( argc != 3 ) {
    cout << "\nUsage: indepset-ip [instance] [root node relax]\n\n";
    cout << "\twhere:" << endl;
    cout << "\t\troot_node_relax = 0: LP relaxation" << endl;
    cout << "\t\troot_node_relax = 1: root node relaxation" << endl;
    cout << endl;
    exit(1);
  }

  bool root_node_relax = (atoi(argv[2]) == 1);

  cout << "\n";
  cout << "-------------------------------------------" << endl;
  cout << "IP - Clique formulation" << endl;
  cout << "-------------------------------------------" << endl;
  cout << "\n";

  // read independent set instance
  IndepSetInst* inst = new IndepSetInst;
  inst->read_DIMACS(argv[1]);

  double clique_time = 0;

  try {

    IloEnv env;
    IloModel model(env);

    IloNumVarArray x(env, inst->graph->n_vertices);

    if ( root_node_relax ) {
      for( int i = 0; i < inst->graph->n_vertices; ++i ) {
        x[i] = IloIntVar(env, 0, 1);
      }
    } else {
      for( int i = 0; i < inst->graph->n_vertices; ++i ) {
        x[i] = IloNumVar(env, 0, 1);
      }
    }
    model.add( IloMaximize(env, IloSum(x)) );


    // -----------------------------------------------
    // Clique decomposition
    // -----------------------------------------------

    stats.start_timer(1);

    clique_decomposition(model, x, inst);

    stats.end_timer(1);
    clique_time = stats.get_time(1);
    cout << "Time to generate cliques: " << clique_time << endl;


    // -----------------------------------------------
    // Clique formulation
    // -----------------------------------------------

    for( int i = 0; i < (int)cliques.size(); ++i ) {
      IloExpr clique_expr(env);
      for( int j = 0; j < (int)cliques[i].size(); ++j ) {
	clique_expr = clique_expr + x[ cliques[i][j] ];
      }
      model.add( clique_expr <= 1 );
    }

    // -----------------------------------------------
    // Edge formulation
    // -----------------------------------------------

    // edge formulation
    //    for( int i = 0; i < inst->graph->n_vertices; ++i ) {
    //      for( int j = i+1; j < inst->graph->n_vertices; ++j ) {
    //        if( inst->graph->is_adj(i, j) ) {
    //          model.add( x[i] + x[j] <= 1 );
    //        }
    //      }
    //    }


    // -----------------------------------------------
    // CPLEX formulation
    // -----------------------------------------------

    stats.start_timer(0);

    IloCplex cplex(model);

    //cplex.exportModel("test.lp");

    // time limit & threads
    cplex.setParam(IloCplex::TiLim, 3600);
    cplex.setParam(IloCplex::Threads, 1);
    cplex.setParam(IloCplex::RootAlg, 4);

    if ( root_node_relax ) {
      cplex.setParam(IloCplex::NodeLim, 0);
    }
    //cplex.exportModel("test.lp");

    cplex.solve();
    stats.end_timer(0);

    double bound = -1;
    if ( root_node_relax ) {
      bound = cplex.getBestObjValue();
    } else {
      bound = cplex.getObjValue();
    }

    cout << "CPLEX obj: " << bound << endl;

    ofstream statsfile("stats.txt", ios::app);
    statsfile << argv[1];
    statsfile << "\t" << root_node_relax;
    statsfile << "\t" << bound;
    statsfile << "\t" << stats.get_time(1);
    statsfile << "\t" << stats.get_time(0);
    statsfile << "\t" << stats.get_time(0) + stats.get_time(1);
    statsfile << "\n";
    statsfile.close();

  } catch( IloException &ex ) {
    cout << "error: " << ex << endl;
    exit(1);
  }

  return 0;

}

